"use strict";

if (typeof (JSIL) === "undefined")
  throw new Error("JSIL.Core is required");

if (!$jsilcore)  
  throw new Error("JSIL.Core is required");

JSIL.MakeClass("System.ValueType", "System.Enum", true, [], function ($) {
  $.ImplementInterfaces(
    /* 0 */ $jsilcore.TypeRef("System.IConvertible")
  );
});

JSIL.ImplementExternals("System.Object", function ($) {
  $.RawMethod(true, "CheckType",
    function (value) {
      var type = typeof (value); 
      return (type=== "object" || type === "number" || type === "string");
    }
  );

  // FIXME: Remove this once the expressions stuff doesn't rely on it anymore
  $.RawMethod(false, "__Initialize__",
    function (initializer) {
      var isInitializer = function (v) {
        return (typeof (v) === "object") && (v !== null) && 
          (
            (Object.getPrototypeOf(v) === JSIL.CollectionInitializer.prototype) ||
            (Object.getPrototypeOf(v) === JSIL.ObjectInitializer.prototype)
          );
      };

      if (JSIL.IsArray(initializer)) {
        JSIL.ApplyCollectionInitializer(this, initializer);
        return this;
      } else if (isInitializer(initializer)) {
        initializer.Apply(this);
        return this;
      }

      for (var key in initializer) {
        if (!initializer.hasOwnProperty(key))
          continue;

        var value = initializer[key];

        if (isInitializer(value)) {
          this[key] = value.Apply(this[key]);
        } else {
          this[key] = value;
        }
      }

      return this;
    }
  );


  $.Method({Static: false, Public: true}, "GetType",
    new JSIL.MethodSignature($jsilcore.TypeRef("System.Type"), [], [], $jsilcore),
    function Object_GetType () {
      return this.__ThisType__;
    }
  );

  $.Method({Static: false, Public: true}, "Object.Equals",
    new JSIL.MethodSignature($.Boolean, [$.Object], [], $jsilcore),
    function Object_Equals (rhs) {
      return this === rhs;
    }
  );

  $.Method({Static:false, Public:true }, "GetHashCode", 
    (new JSIL.MethodSignature($.Int32, [], [])), 
    function Object_GetHashCode () {
      return JSIL.HashCodeInternal(this);
    }
  );

  // HACK: Prevent infinite recursion
  var currentMemberwiseCloneInvocation = null;

  $.Method({Static: false, Public: false}, "MemberwiseClone",
    new JSIL.MethodSignature($.Object, [], [], $jsilcore),
    function Object_MemberwiseClone () {
      var result = null;

      // HACK: Handle Object.MemberwiseClone direct invocation
      if (currentMemberwiseCloneInvocation === this.MemberwiseClone) {
        result = new System.Object();
      } else {
        currentMemberwiseCloneInvocation = this.MemberwiseClone;
        try {
          result = this.MemberwiseClone();
        } finally {
          currentMemberwiseCloneInvocation = null;
        }
      }

      return result;
    }
  );

  $.Method({Static: false, Public: true}, ".ctor",
    new JSIL.MethodSignature(null, []),
    function Object__ctor () {
    }
  );

  $.Method({Static: false, Public: true}, "toString",
    new JSIL.MethodSignature($.String, [], [], $jsilcore),
    function Object_ToString () {
      return JSIL.GetTypeName(this);
    }
  );

  $.Method({Static:true , Public:true }, "ReferenceEquals", 
    (new JSIL.MethodSignature($.Boolean, [$.Object, $.Object], [])), 
    function ReferenceEquals (objA, objB) {
      return objA === objB;
    }
  );

});

JSIL.MakeClass(Object, "System.Object", true, [], function ($) {
  $.ExternalMethod({Static: false, Public: true}, ".ctor",
    new JSIL.MethodSignature(null, [], [], $jsilcore)
  );

  $.ExternalMethod({Static: false, Public: true}, "GetType",
    new JSIL.MethodSignature("System.Type", [], [], $jsilcore)
  );

  $.ExternalMethod({Static: false, Public: true}, "Object.Equals",
    new JSIL.MethodSignature("System.Boolean", [$.Type], [], $jsilcore)
  );

  $.ExternalMethod({Static: false, Public: true}, "MemberwiseClone",
    new JSIL.MethodSignature("System.Object", [], [], $jsilcore)
  );

  $.ExternalMethod({Static: false, Public: true}, "toString",
    new JSIL.MethodSignature("System.String", [], [], $jsilcore)
  );

  $jsilcore.SystemObjectInitialized = true;
});

JSIL.MakeClass("System.Object", "JSIL.CollectionInitializer", true, [], function ($) {
  $.RawMethod(false, ".ctor",
    function () {
      this.values = Array.prototype.slice.call(arguments);
    }
  );

  $.RawMethod(false, "Apply",
    function (previousValue) {
      JSIL.ApplyCollectionInitializer(previousValue, this.values);

      return previousValue;
    }
  );
});

JSIL.MakeClass("System.Object", "JSIL.ObjectInitializer", true, [], function ($) {
  $.RawMethod(false, ".ctor",
    function (newInstance, initializer) {
      this.hasInstance = (newInstance !== null);
      this.instance = newInstance;
      this.initializer = initializer;
    }
  );

  $.RawMethod(false, "Apply",
    function (previousValue) {
      var result = this.hasInstance ? this.instance : previousValue;

      if (result)
        result.__Initialize__(this.initializer);
      else
        JSIL.Host.warning("Object initializer applied to null/undefined!");

      return result;
    }
  );
});

JSIL.MakeClass("System.Object", "System.ValueType", true, [], function ($) {
  $.Method({Static: false, Public: true}, "Object.Equals",
    new JSIL.MethodSignature(System.Boolean, [System.Object]),
    function (rhs) {
      return JSIL.StructEquals(this, rhs);
    }
  );
});

JSIL.MakeInterface(
  "System.IDisposable", true, [], function ($) {
    $.Method({}, "Dispose", (JSIL.MethodSignature.Void));
  }, []);

JSIL.MakeInterface(
  "System.IEquatable`1", true, ["T"], function ($) {
    $.Method({}, "Equals", (new JSIL.MethodSignature($jsilcore.TypeRef("System.Boolean"), [new JSIL.GenericParameter("T", "System.IEquatable`1")], [])));
  }, []);

JSIL.MakeInterface(
  "System.Collections.IEnumerator", true, [], function ($) {
    $.Method({}, "MoveNext", (new JSIL.MethodSignature($jsilcore.TypeRef("System.Boolean"), [], [])));
    $.Method({}, "get_Current", (new JSIL.MethodSignature($jsilcore.TypeRef("System.Object"), [], [])));
    $.Method({}, "Reset", (JSIL.MethodSignature.Void));
    $.Property({}, "Current");
  }, []);

JSIL.MakeInterface(
  "System.Collections.IDictionaryEnumerator", true, [], function ($) {
    $.Method({}, "get_Key", new JSIL.MethodSignature($.Object, [], []));
    $.Method({}, "get_Value", new JSIL.MethodSignature($.Object, [], []));
    // FIXME
    // $.Method({}, "get_Entry", new JSIL.MethodSignature($jsilcore.TypeRef("System.Collections.DictionaryEntry"), [], []));
    $.Property({}, "Key");
    $.Property({}, "Value");
    $.Property({}, "Entry");
  }, [$jsilcore.TypeRef("System.Collections.IEnumerator")]);

JSIL.MakeInterface(
  "System.Collections.IEnumerable", true, [], function ($) {
    $.Method({}, "GetEnumerator", (new JSIL.MethodSignature($jsilcore.TypeRef("System.Collections.IEnumerator"), [], [])));
  }, []);

JSIL.MakeInterface(
  "System.Collections.Generic.IEnumerator`1", true, ["out T"], function ($) {
    $.Method({}, "get_Current", (new JSIL.MethodSignature(new JSIL.GenericParameter("T", "System.Collections.Generic.IEnumerator`1"), [], [])));
    $.Property({}, "Current");
  }, [$jsilcore.TypeRef("System.IDisposable"), $jsilcore.TypeRef("System.Collections.IEnumerator")]);

JSIL.MakeInterface(
  "System.Collections.Generic.IEnumerable`1", true, ["out T"], function ($) {
    $.Method({}, "GetEnumerator", (new JSIL.MethodSignature($jsilcore.TypeRef("System.Collections.Generic.IEnumerator`1", [new JSIL.GenericParameter("T", "System.Collections.Generic.IEnumerable`1")]), [], [])));
  }, [$jsilcore.TypeRef("System.Collections.IEnumerable")]);




/* delegate System.Comparison`1 */

JSIL.DeclareNamespace("System");
JSIL.MakeDelegate("System.Comparison`1", true, ["in T"], new JSIL.MethodSignature($jsilcore.TypeRef("System.Int32"), [new JSIL.GenericParameter("T", "System.Comparison`1").in(), new JSIL.GenericParameter("T", "System.Comparison`1").in()]));

/* interface System.Linq.IOrderedEnumerable`1 */

JSIL.DeclareNamespace("System.Linq");
JSIL.MakeInterface(
  "System.Linq.IOrderedEnumerable`1", true, ["TElement"], function ($) {
      //Note: the code commented below was to implement the IOrderedEnumerable.CreateOrderedEnumerable method but I can't seem to make it work properly so we'll remove it for now.
      //$.Method({}, "CreateOrderedEnumerable", new JSIL.MethodSignature($jsilcore.TypeRef("System.Linq.IOrderedEnumerable`1", [$.GenericParameter("TElement")]), [
      //    $jsilcore.TypeRef("System.Func`2", [$.GenericParameter("TElement"), "!!0"]), $jsilcore.TypeRef("System.Collections.Generic.IComparer`1", ["!!0"]),
      //    $.Boolean
      //], ["TKey"]));
  }, [$jsilcore.TypeRef("System.Collections.Generic.IEnumerable`1", [new JSIL.GenericParameter("TElement", "System.Linq.IOrderedEnumerable`1")]), $jsilcore.TypeRef("System.Collections.IEnumerable")]);


/* class System.Linq.OrderedEnumerable`2 */

(function OrderedEnumerable$b2$Members() {
    var $, $thisType;

    var $T00 = function () {
        return ($T00 = JSIL.Memoize($jsilcore.System.Boolean))();
    };
    var $T01 = function () {
        return ($T01 = JSIL.Memoize(System.Linq.Enumerable))();
    };
    //function OrderedEnumerable$b2__ctor(collection, keySelector, comparer, descending) {
    //    var $s00 = new JSIL.MethodSignature(null, [$jsilcore.TypeRef("System.Collections.Generic.IEnumerable`1", [$thisType.TElement.get(this)])]);
    //    var $s01 = new JSIL.MethodSignature(null, [$jsilcore.TypeRef("System.Comparison`1", [$thisType.TElement.get(this)])]);
    //    $s00.Call($jsilcore.System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "_ctor", null, this, collection);
    //    this._keySelector = keySelector;
    //    this._comparer = comparer;
    //    if (this._comparer !== null) {
    //        $s01.Call($jsilcore.System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, this, $jsilcore.System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['CustomComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "CustomComparer", new JSIL.MethodSignature($jsilcore.System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
    //    } else {
    //        $s01.Call($jsilcore.System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, this, $jsilcore.System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['DefaultComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "DefaultComparer", new JSIL.MethodSignature($jsilcore.System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
    //    }
    //    if (descending) {
    //        $jsilcore.System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['Reverse'].call(this);
    //    }
    //};

    function OrderedEnumerable$b2__ctor(collection, keySelector, comparer, descending) {
        var $temp00;
        var $s00 = new JSIL.ConstructorSignature($jsilcore.TypeRef("System.Collections.Generic.List`1", [$thisType.TElement.get(this)]), null);
        var $s01 = new JSIL.MethodSignature(null, [$jsilcore.TypeRef("System.Comparison`1", [$thisType.TElement.get(this)])]);
        var $im00 = System.Collections.Generic.IComparer$b1.Of($thisType.TKey.get(this)).Compare;
        System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['_ctor'].call(this);
        this._keySelector = keySelector;
        this._comparer = comparer;
        if ($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).$Is(collection)) {
            var collectionAsIOrderedEnumerable = $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).$Cast(collection);
            var parentComparer = collectionAsIOrderedEnumerable._comparer;
            var parentKeySelector = collectionAsIOrderedEnumerable._keySelector;
            if (parentComparer === null) {
                parentComparer = System.Collections.Generic.Comparer$b1.Of($thisType.TKey.get(this))['get_Default']();
            }
            var lastValue = keySelector(JSIL.CloneParameter($thisType.TElement.get(this), $T01()['ElementAt$b1']($thisType.TElement.get(this))(collectionAsIOrderedEnumerable, 0)));
            var tempList = null;
            var parentKeyInTempList = (
              $thisType.TKey.get(this).IsValueType
                 ? JSIL['CreateInstanceOfType']($thisType.TKey.get(this))
                 : null)
            ;

            for (var a$0 = collectionAsIOrderedEnumerable._items, i$0 = 0, l$0 = (collectionAsIOrderedEnumerable._size | 0) ; i$0 < l$0; ($temp00 = i$0,
                i$0 = ((i$0 + 1) | 0),
                $temp00)) {
                var testedElement = JSIL.CloneParameter($thisType.TElement.get(this), a$0[i$0]);
                var parentKey = JSIL.CloneParameter($thisType.TKey.get(this), parentKeySelector(JSIL.CloneParameter($thisType.TElement.get(this), testedElement)));
                if (tempList === null) {
                    tempList = $s00.Construct();
                    tempList['Add'](JSIL.CloneParameter($thisType.TElement.get(this), testedElement));
                    parentKeyInTempList = parentKey;
                } else {
                    var parentComparison = ($im00.Call(parentComparer, null, JSIL.CloneParameter($thisType.TKey.get(this), parentKeyInTempList), JSIL.CloneParameter($thisType.TKey.get(this), parentKey)) | 0);
                    if (parentComparison !== 0) {
                        if (this._comparer !== null) {
                            $s01.Call(System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, tempList, System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['CustomComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "CustomComparer", new JSIL.MethodSignature(System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
                        } else {
                            $s01.Call(System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, tempList, System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['DefaultComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "DefaultComparer", new JSIL.MethodSignature(System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
                        }
                        if (descending) {
                            System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['Reverse'].call(tempList);
                        }
                        System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['AddRange'].call(this, tempList);
                        parentKeyInTempList = parentKey;
                        tempList['Clear']();
                        tempList['Add'](JSIL.CloneParameter($thisType.TElement.get(this), testedElement));
                    } else {
                        tempList['Add'](JSIL.CloneParameter($thisType.TElement.get(this), testedElement));
                    }
                }
            }
            if (this._comparer !== null) {
                $s01.Call(System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, tempList, System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['CustomComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "CustomComparer", new JSIL.MethodSignature(System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
            } else {
                $s01.Call(System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, tempList, System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['DefaultComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "DefaultComparer", new JSIL.MethodSignature(System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
            }
            if (descending) {
                System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['Reverse'].call(tempList);
            }
            System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['AddRange'].call(this, tempList);
        } else {
            System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['AddRange'].call(this, collection);

            if (this._comparer !== null) {
                $s01.Call(System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, this, System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['CustomComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "CustomComparer", new JSIL.MethodSignature(System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
            } else {
                $s01.Call(System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype, "Sort", null, this, System.Comparison$b1.Of($thisType.TElement.get(this))['New'](this, $thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)).prototype['DefaultComparer'], function () { return JSIL.GetMethodInfo($thisType.Of($thisType.TElement.get(this), $thisType.TKey.get(this)), "DefaultComparer", new JSIL.MethodSignature(System.Int32, [$thisType.TElement.get(this), $thisType.TElement.get(this)]), false); }));
            }
            if (descending) {
                System.Collections.Generic.List$b1.Of($thisType.TElement.get(this)).prototype['Reverse'].call(this);
            }
        }
    };




    function OrderedEnumerable$b2_CustomComparer(a, b) {
        return this._comparer.Compare(JSIL.CloneParameter($thisType.TKey.get(this), this._keySelector(a)),
          JSIL.CloneParameter($thisType.TKey.get(this), this._keySelector(b)));
    };

    function OrderedEnumerable$b2_DefaultComparer(a, b) {
        return $jsilcore.System.Collections.Generic.Comparer$b1.Of($thisType.TKey.get(this))['get_Default']()['Compare'](
          JSIL.CloneParameter($thisType.TKey.get(this), this._keySelector(a)),
          JSIL.CloneParameter($thisType.TKey.get(this), this._keySelector(b))
        );
    };

    JSIL.MakeType({
        BaseType: $jsilcore.TypeRef("System.Collections.Generic.List`1", [new JSIL.GenericParameter("TElement", "System.Linq.OrderedEnumerable`2")]),
        Name: "System.Linq.OrderedEnumerable`2",
        IsPublic: true,
        IsReferenceType: true,
        GenericParameters: ["TElement", "TKey"],
        MaximumConstructorArguments: 4,
    }, function ($interfaceBuilder) {
        $ = $interfaceBuilder;

        $.Method({ Static: false, Public: true }, ".ctor",
          new JSIL.MethodSignature(null, [
              $jsilcore.TypeRef("System.Collections.Generic.IEnumerable`1", [$.GenericParameter("TElement")]), $jsilcore.TypeRef("System.Func`2", [$.GenericParameter("TElement"), $.GenericParameter("TKey")]),
              $jsilcore.TypeRef("System.Collections.Generic.IComparer`1", [$.GenericParameter("TKey")]), $.Boolean
          ]),
          OrderedEnumerable$b2__ctor
        );

        //Note: the code commented below was to implement the IOrderedEnumerable.CreateOrderedEnumerable method but I can't seem to make it work properly so we'll remove it for now.
    //    $.Method({ Static: false, Public: false, Virtual: true }, "System.Linq.IOrderedEnumerable$lTElement$g.CreateOrderedEnumerable",
    //  new JSIL.MethodSignature($jsilcore.TypeRef("System.Linq.IOrderedEnumerable`1", [$.GenericParameter("TElement")]), [
    //      $jsilcore.TypeRef("System.Func`2", [$.GenericParameter("TElement"), "!!0"]), $jsilcore.TypeRef("System.Collections.Generic.IComparer`1", ["!!0"]),
    //      $.Boolean
    //  ], ["TKey"]),
    //  OrderedEnumerable$b1_System_Linq_IOrderedEnumerable$lTElement$g_CreateOrderedEnumerable$b1
    //)
    //  .Overrides($jsilcore.TypeRef("System.Linq.IOrderedEnumerable`1", [$.GenericParameter("TElement")]), "CreateOrderedEnumerable");

    //    function OrderedEnumerable$b1_System_Linq_IOrderedEnumerable$lTElement$g_CreateOrderedEnumerable$b1(TKey, keySelector, comparer, descending) {
    //        var result = new (System.Linq.OrderedEnumerable$b2.Of($thisType.TElement.get(this), TKey))(this.source, keySelector, comparer, descending);
    //        result.parent = this;
    //        return result;
    //    };

        $.Method({ Static: false, Public: false }, "CustomComparer",
          new JSIL.MethodSignature($.Int32, [$.GenericParameter("TElement"), $.GenericParameter("TElement")]),
          OrderedEnumerable$b2_CustomComparer
        );

        $.Method({ Static: false, Public: false }, "DefaultComparer",
          new JSIL.MethodSignature($.Int32, [$.GenericParameter("TElement"), $.GenericParameter("TElement")]),
          OrderedEnumerable$b2_DefaultComparer
        );

        $['Field']({ Static: false, Public: false }, "_keySelector", $jsilcore.TypeRef("System.Func`2", [$.GenericParameter("TElement"), $.GenericParameter("TKey")]));
        $['Field']({ Static: false, Public: false }, "_comparer", $jsilcore.TypeRef("System.Collections.Generic.IComparer`1", [$.GenericParameter("TKey")]));
        $.ImplementInterfaces(
          /* 8 */ $jsilcore.TypeRef("System.Linq.IOrderedEnumerable`1", [$.GenericParameter("TElement")]),
          /* 9 */ $jsilcore.TypeRef("System.Collections.Generic.IEnumerable`1", [$.GenericParameter("TElement")])
        );

        return function (newThisType) { $thisType = newThisType; };
    });

})();



JSIL.MakeInterface(
  "System.Collections.ICollection", true, [], function ($) {
    $.Method({}, "CopyTo", (new JSIL.MethodSignature(null, [$jsilcore.TypeRef("System.Array"), $.Int32], [])));
    $.Method({}, "get_Count", (new JSIL.MethodSignature($.Int32, [], [])));
    $.Method({}, "get_SyncRoot", (new JSIL.MethodSignature($.Object, [], [])));
    $.Method({}, "get_IsSynchronized", (new JSIL.MethodSignature($.Boolean, [], [])));
    $.Property({}, "Count");
    $.Property({}, "SyncRoot");
    $.Property({}, "IsSynchronized");
  }, [$jsilcore.TypeRef("System.Collections.IEnumerable")]);

JSIL.MakeInterface(
  "System.Collections.IList", true, [], function ($) {
    $.Method({}, "get_Item", (new JSIL.MethodSignature($.Object, [$.Int32], [])));
    $.Method({}, "set_Item", (new JSIL.MethodSignature(null, [$.Int32, $.Object], [])));
    $.Method({}, "Add", (new JSIL.MethodSignature($.Int32, [$.Object], [])));
    $.Method({}, "Contains", (new JSIL.MethodSignature($.Boolean, [$.Object], [])));
    $.Method({}, "Clear", (JSIL.MethodSignature.Void));
    $.Method({}, "get_IsReadOnly", (new JSIL.MethodSignature($.Boolean, [], [])));
    $.Method({}, "get_IsFixedSize", (new JSIL.MethodSignature($.Boolean, [], [])));
    $.Method({}, "IndexOf", (new JSIL.MethodSignature($.Int32, [$.Object], [])));
    $.Method({}, "Insert", (new JSIL.MethodSignature(null, [$.Int32, $.Object], [])));
    $.Method({}, "Remove", (new JSIL.MethodSignature(null, [$.Object], [])));
    $.Method({}, "RemoveAt", (new JSIL.MethodSignature(null, [$.Int32], [])));
    $.Property({}, "Item");
    $.Property({}, "IsReadOnly");
    $.Property({}, "IsFixedSize");
  }, [$jsilcore.TypeRef("System.Collections.ICollection"), $jsilcore.TypeRef("System.Collections.IEnumerable")]);

JSIL.MakeInterface(
  "System.Collections.Generic.ICollection`1", true, ["T"], function ($) {
    $.Method({}, "get_Count", (new JSIL.MethodSignature($.Int32, [], [])));
    $.Method({}, "get_IsReadOnly", (new JSIL.MethodSignature($.Boolean, [], [])));
    $.Method({}, "Add", (new JSIL.MethodSignature(null, [new JSIL.GenericParameter("T", "System.Collections.Generic.ICollection`1")], [])));
    $.Method({}, "Clear", (JSIL.MethodSignature.Void));
    $.Method({}, "Contains", (new JSIL.MethodSignature($.Boolean, [new JSIL.GenericParameter("T", "System.Collections.Generic.ICollection`1")], [])));
    $.Method({}, "CopyTo", (new JSIL.MethodSignature(null, [$jsilcore.TypeRef("System.Array", [new JSIL.GenericParameter("T", "System.Collections.Generic.ICollection`1")]), $.Int32], [])));
    $.Method({}, "Remove", (new JSIL.MethodSignature($.Boolean, [new JSIL.GenericParameter("T", "System.Collections.Generic.ICollection`1")], [])));
    $.Property({}, "Count");
    $.Property({}, "IsReadOnly");
  }, [$jsilcore.TypeRef("System.Collections.Generic.IEnumerable`1", [new JSIL.GenericParameter("T", "System.Collections.Generic.ICollection`1")]), $jsilcore.TypeRef("System.Collections.IEnumerable")]);

JSIL.MakeInterface(
  "System.Collections.Generic.IList`1", true, ["T"], function ($) {
    $.Method({}, "get_Item", (new JSIL.MethodSignature(new JSIL.GenericParameter("T", "System.Collections.Generic.IList`1"), [$.Int32], [])));
    $.Method({}, "set_Item", (new JSIL.MethodSignature(null, [$.Int32, new JSIL.GenericParameter("T", "System.Collections.Generic.IList`1")], [])));
    $.Method({}, "IndexOf", (new JSIL.MethodSignature($.Int32, [new JSIL.GenericParameter("T", "System.Collections.Generic.IList`1")], [])));
    $.Method({}, "Insert", (new JSIL.MethodSignature(null, [$.Int32, new JSIL.GenericParameter("T", "System.Collections.Generic.IList`1")], [])));
    $.Method({}, "RemoveAt", (new JSIL.MethodSignature(null, [$.Int32], [])));
    $.Property({}, "Item");
  }, [$jsilcore.TypeRef("System.Collections.Generic.ICollection`1", [new JSIL.GenericParameter("T", "System.Collections.Generic.IList`1")]), $jsilcore.TypeRef("System.Collections.Generic.IEnumerable`1", [new JSIL.GenericParameter("T", "System.Collections.Generic.IList`1")]), $jsilcore.TypeRef("System.Collections.IEnumerable")]);

JSIL.ImplementExternals("System.Array", function ($) {
  $.RawMethod(true, "CheckType", JSIL.IsSystemArray);

  $.RawMethod(true, "Of", function Array_Of () {
    // Ensure System.Array is initialized.
    var _unused = $jsilcore.System.Array.Of;

    return $jsilcore.ArrayOf.apply(null, arguments);
  });
});
  
JSIL.MakeClass("System.Object", "System.Array", true, [], function ($) {
  $.SetValue("__IsArray__", true);

  $.RawMethod(false, "GetLength", function () {
    return this.length;
  });
  $.RawMethod(false, "GetLowerBound", function () {
    return 0;
  });
  $.RawMethod(false, "GetUpperBound", function () {
    return this.length - 1;
  });

  var typeObject = $.typeObject;
  var publicInterface = $.publicInterface;
  var types = {};

  var checkType = function Array_CheckType (value) {
    return JSIL.IsSystemArray(value);
  };

  $.RawMethod(true, "CheckType", checkType);

  var of = function Array_Of (elementType) {
    if (typeof (elementType) === "undefined")
      JSIL.RuntimeError("Attempting to create an array of an undefined type");

    var _ = JSIL.ResolveTypeReference(elementType);
    var elementTypePublicInterface = _[0];
    var elementTypeObject = _[1];

    var elementTypeId = elementTypeObject.__TypeId__;
    if (typeof (elementTypeId) === "undefined")
      JSIL.RuntimeError("Element type missing type ID");

    var compositePublicInterface = types[elementTypeObject.__TypeId__];

    if (typeof (compositePublicInterface) === "undefined") {
      var typeName = elementTypeObject.__FullName__ + "[]";

      var compositeTypeObject = JSIL.CreateDictionaryObject(typeObject);
      compositePublicInterface = function (size) {
        JSIL.RuntimeError("Invalid use of Array constructor. Use JSIL.Array.New.");
      };
      compositePublicInterface.prototype = JSIL.CreatePrototypeObject(publicInterface.prototype);

      compositePublicInterface.__Type__ = compositeTypeObject;
      JSIL.SetTypeId(
        compositeTypeObject, compositePublicInterface, 
        typeObject.__TypeId__ + "[" + elementTypeObject.__TypeId__ + "]"
      );
      compositePublicInterface.CheckType = publicInterface.CheckType;

      JSIL.SetValueProperty(compositeTypeObject, "__PublicInterface__", compositePublicInterface);
      JSIL.SetValueProperty(
        compositeTypeObject, "__FullName__", 
        compositeTypeObject.__FullNameWithoutArguments__ = typeName
      );
      JSIL.SetValueProperty(compositeTypeObject, "__IsReferenceType__", true);
      compositeTypeObject.__IsArray__ = true;
      compositeTypeObject.__ElementType__ = elementTypeObject;
      compositeTypeObject.__IsClosed__ = Object.getPrototypeOf(compositeTypeObject.__ElementType__) !== JSIL.GenericParameter.prototype;

      JSIL.SetValueProperty(compositePublicInterface, "CheckType", checkType);
      JSIL.SetValueProperty(compositeTypeObject, "toString", function ArrayType_ToString () {
        return typeName;
      });

      compositePublicInterface.prototype = JSIL.MakeProto(
        publicInterface, compositeTypeObject, typeName, true, elementTypeObject.__Context__
      );
      JSIL.SetValueProperty(compositePublicInterface, "toString", function ArrayPublicInterface_ToString () {
        return "<" + typeName + " Public Interface>";
      });

      JSIL.MakeCastMethods(compositePublicInterface, compositeTypeObject, "array");

      types[elementTypeObject.__TypeId__] = compositePublicInterface;
    }

    return compositePublicInterface;
  };

  $jsilcore.ArrayOf = of;

  $.RawMethod(true, "Of$NoInitialize", of);
  $.RawMethod(true, "Of", of);
});

JSIL.MakeClass("System.Array", "JSIL.MultidimensionalArray", true, [], function ($) {
  $.Method({Static: false, Public: true }, ".ctor",
    new JSIL.MethodSignature(null, ["System.Type", "System.Array", "System.Array"], [], $jsilcore),
    function (type, dimensions, initializer) {
      if (dimensions.length < 2)
        throw new Error("Must have at least two dimensions: " + String(dimensions));

      var totalSize = dimensions[0];
      for (var i = 1; i < dimensions.length; i++)
        totalSize *= dimensions[i];

      this._type = type;
      this._dimensions = dimensions;
      var items = this._items = JSIL.Array.New(type, totalSize);

      JSIL.SetValueProperty(this, "length", totalSize);

      if (JSIL.IsArray(initializer)) {
        JSIL.Array.ShallowCopy(items, initializer);
      } else {
        JSIL.Array.Erase(items, type);
      }

      switch (dimensions.length) {
        case 2:
          var height = this.length0 = dimensions[0];
          var width = this.length1 = dimensions[1];

          JSIL.SetValueProperty(this, "Get", 
            function Get (y, x) {
              return items[(y * width) + x];
            }
          );
          JSIL.SetValueProperty(this, "GetReference", 
            function GetReference (y, x) {
              return new JSIL.MemberReference(items, (y * width) + x);
            }
          );
          JSIL.SetValueProperty(this, "Set", 
            function Set (y, x, value) {
              return items[(y * width) + x] = value;
            }
          );
          JSIL.SetValueProperty(this, "GetLength", 
            function GetLength (i) {
              return dimensions[i];
            }
          );
          JSIL.SetValueProperty(this, "GetUpperBound", 
            function GetUpperBound (i) {
              return dimensions[i] - 1;
            }
          );
          break;
        case 3:
          var depth = this.length0 = dimensions[0];
          var height = this.length1 = dimensions[1];
          var width = this.length2 = dimensions[2];
          var heightxwidth = height * width;

          JSIL.SetValueProperty(this, "Get", 
            function Get (z, y, x) {
              return items[(z * heightxwidth) + (y * width) + x];      
            }
          );
          JSIL.SetValueProperty(this, "GetReference", 
            function GetReference (z, y, x) {
              return new JSIL.MemberReference(items, (z * heightxwidth) + (y * width) + x);
            }
          );
          JSIL.SetValueProperty(this, "Set", 
            function Set (z, y, x, value) {
              return items[(z * heightxwidth) + (y * width) + x] = value;
            }
          );
          JSIL.SetValueProperty(this, "GetLength", 
            function GetLength (i) {
              return dimensions[i];
            }
          );
          JSIL.SetValueProperty(this, "GetUpperBound", 
            function GetUpperBound (i) {
              return dimensions[i] - 1;
            }
          );
          break;
      }
    }
  );

  $.RawMethod(true, "New",
    function (type) {
      var initializer = arguments[arguments.length - 1];
      var numDimensions = arguments.length - 1;

      if (JSIL.IsArray(initializer))
        numDimensions -= 1;
      else
        initializer = null;

      if (numDimensions < 1)
        throw new Error("Must provide at least one dimension");
      else if ((numDimensions == 1) && (initializer === null))
        return System.Array.New(type, arguments[1]);

      var dimensions = Array.prototype.slice.call(arguments, 1, 1 + numDimensions);

      return new JSIL.MultidimensionalArray(type, dimensions, initializer);
    }
  );

  $.SetValue("__IsArray__", true);
});

JSIL.ImplementExternals(
  "System.Array", function ($) {
    $.Method({ Static: true, Public: true }, "Resize",
      new JSIL.MethodSignature(null, [$jsilcore.TypeRef("JSIL.Reference", [$jsilcore.TypeRef("System.Array", ["!!0"])]), $.Int32], ["T"]),
      function (type, arr, newSize) {
        var oldArray = arr.get(), newArray = null;
        var oldLength = oldArray.length;

        if (Array.isArray(oldArray)) {
          newArray = oldArray;
          newArray.length = newSize;

          for (var i = oldLength; i < newSize; i++)
            newArray[i] = JSIL.DefaultValue(type);
        } else {
          newArray = JSIL.Array.New(type, newSize);

          JSIL.Array.CopyTo(oldArray, newArray, 0);
        }

        arr.set(newArray);
      }
    );
  }
);

JSIL.ImplementExternals("System.Enum", function ($) {
  $.Method({Static:true , Public:true }, "ToObject",
    (new JSIL.MethodSignature($.Object, ["System.Type", $.Int32], [])),
    function ToObject (enumType, value) {
      return enumType[enumType.__ValueToName__[value]];
    }
  );

  $.Method({Static:false, Public:false, Virtual:true }, "ToInt32",
    new JSIL.MethodSignature($.Int32, [$jsilcore.TypeRef("System.IFormatProvider")], []),
    function (provider) {
      return $jsilcore.System.Convert.ToInt32(this.value, provider);
    }
  );

  $.Method({Static:false, Public:false, Virtual:true }, "ToInt64",
    new JSIL.MethodSignature($.Int64, [$jsilcore.TypeRef("System.IFormatProvider")], []),
    function (provider) {
      return $jsilcore.System.Convert.ToInt64(this.value, provider);
    }
  );

  $.Method({Static: false, Public: true}, "Object.Equals",
    new JSIL.MethodSignature(System.Boolean, [System.Object]),
    function (rhs) {
      if (rhs === null)
        return false;
      
      return (this.__ThisType__ === rhs.__ThisType__) &&
        (this.value === rhs.value);
    }
  );
});

JSIL.MakeClass($jsilcore.TypeRef("System.Object"), "System.Attribute", true, [], function ($) {
  $.Method({Static: false, Public: true }, ".ctor",
    (JSIL.MethodSignature.Void),
    function () {
    }
  );
});

JSIL.MakeEnum(
  "System.TypeCode", true, {
    Empty: 0, 
    Object: 1, 
    DBNull: 2, 
    Boolean: 3, 
    Char: 4, 
    SByte: 5, 
    Byte: 6, 
    Int16: 7, 
    UInt16: 8, 
    Int32: 9, 
    UInt32: 10, 
    Int64: 11, 
    UInt64: 12, 
    Single: 13, 
    Double: 14, 
    Decimal: 15, 
    DateTime: 16, 
    String: 18
  }, false
);
